import pickle
from base64 import b64decode, b64encode
from pathlib import Path
from time import time
from typing import List, Union

import intake
import numpy
from intake.container.persist import PersistStore
from intake_parquet import ParquetSource
import dask.dataframe as dd
from pandas import Series


def make_meta(col):
    return Series(dtype=numpy.object)


# TODO(higumachan): のちのちに辛くなってくる可能性があるので可能ならばJSONで書き出したりができるようにしておきたい
def _dump_with_pickle(obj):
    s = pickle.dumps(obj)
    return b64encode(s).decode("ascii")


def _load_with_pickle(pickle_str: str):
    b = b64decode(pickle_str.encode("ascii"))
    return pickle.loads(b)


class UserDefinedClassParquetSource(ParquetSource):
    """
    originalのPython classが入っていたdask dataframeをpaquetから読み込むときに使うSource
    """
    def to_spark(self):
        raise NotImplementedError("only use with dask")

    def _to_dask(self):
        super()._to_dask()
        for col in self.metadata["user_defined_class_columns"]:
            self._df[col] = self._df[col].map(_load_with_pickle, meta=make_meta(col))
        return self._df

    @classmethod
    def _from_parqeuet_source(cls, ps: ParquetSource):
        ccps = cls.__new__(cls)
        ccps.__dict__ = ps.__dict__
        return ccps

    def persist(self, ttl=None, **kwargs):
        if self.container != 'dataframe':
            raise ValueError(f"{self.__class__.classname} dataframe only")
        store = PersistStore()
        path = store.getdir(self)
        out = self.export(path, **kwargs)
        out.metadata.update({
            "ttl": ttl,
            'cat': {} if self.cat is None else self.cat.__getstate__(),
        })
        from dask.base import tokenize
        tok = tokenize(self)
        store.add(tok, out)
        return out

    def export(self, path, **kwargs):
        df = self.to_dask()
        out = upload_with_user_defined_class(df, path, self.metadata["user_defined_class_columns"], **kwargs)
        from dask.base import tokenize
        tok = tokenize(self)
        metadata = {'timestamp': time(),
                    'original_metadata': self.metadata,
                    'original_source': self.__getstate__(),
                    'original_name': self.name,
                    'original_tok': tok,
                    'persist_kwargs': kwargs,}
        out.metadata.update(metadata)
        out.name = self.name
        return out


def upload_with_user_defined_class(df: dd.DataFrame, path: Union[str, Path], user_defined_class_columns: List[str], **kwargs) -> UserDefinedClassParquetSource:
    """
    originalのPython classが入っているdask dataframeをparquet形式でintakeにuploadするときに使う関数。
    intake.uploadの代わりに利用することができます。
    :param df: 
    :param path: 
    :param user_defined_class_columns: originalのPython classが入っている列の名前のリスト
    :param kwargs: pass intake.upload
    :return: 
    """
    df = df.copy()
    for col in user_defined_class_columns:
        df[col] = df[col].map(_dump_with_pickle, meta=make_meta(col))
    parquet_source: ParquetSource = intake.upload(df, str(path), **kwargs)
    parquet_source.metadata.update({"user_defined_class_columns": user_defined_class_columns})
    return UserDefinedClassParquetSource._from_parqeuet_source(parquet_source)



from collections import OrderedDict
from typing import Any, List

import pandas
from dask.dataframe import DataFrame

from framania.daskmania.meta import make_meta, map_partitions_as_meta
from framania.pandasmania.stack import stack_list_column as pandas_stack_list_column
from framania.pandasmania.stack import stack_list_columns as pandas_stack_list_columns
from framania.pandasmania.stack import stack_dict_column as pandas_stack_dict_column
from framania.pandasmania.stack import stack_columns as pandas_stack_columns


def stack_list_column(dd: DataFrame, list_column: Any, output_dtype: Any, keep_columns: List[Any] = None):
    """
    API to create stack dask dataframe from column containing list-like objects.

    Args:
        df (dask.dataframe.DataFrame): target dask dataframe
        list_column (Any): name of column containing list-like objects.
        output_dtype (Any): dtype of output column. may depends on contents of list.
        keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
    Returns:
        result dask dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': ['1,2,3', '2,3,4', '3,4'], 'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
        >>> pd['a-list'] = pd['a'].str.split(',')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
                 a  b     a-list
        idx
        0    1,2,3  1  [1, 2, 3]
        1    2,3,4  2  [2, 3, 4]
        2      3,4  3     [3, 4]
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> result = stack_list_column(dd, 'a-list', 'str', ['b'])
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
             b a-list
        idx
        0    1      1
        0    1      2
        0    1      3
        1    2      2
        1    2      3
        1    2      4
        2    3      3
        2    3      4
    """
    meta = make_meta((dd.index.name, dd.index.dtype),
                     [(k_column, dd[k_column].dtype) for k_column in keep_columns] +
                     [(list_column, output_dtype)])
    return dd.map_partitions(lambda pd: pandas_stack_list_column(pd, list_column, output_dtype, keep_columns),
                             meta=meta)


def stack_list_columns(dd: DataFrame, list_columns: List[Any], output_dtypes: List[Any],
                       keep_columns: List[Any] = None):
    """
        API to create stack dask dataframe from columns containing list-like objects.
        List elements which has same list index will be assigned to same row.
        If length of lists are not same, NaN will be filled automatically.

        Args:
            df (pandas.DataFrame): target dask dataframe
            list_columns (List[Any]): name of columns containing list-like objects.
            output_dtypes (List[Any]): dtypes of output columns. may depends on contents of list.
            keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
        Returns:
            result dask dataframe
        Examples:
            >>> import dask.dataframe
            >>> import pandas
            >>> pd = pandas.DataFrame({'a1': ['1,2,3', '2,3,4', '3,4'], 'a2': ['a,b,c', 'b,c,d', 'c,d'],
            ...          'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
            >>> pd['a1-list'] = pd['a1'].str.split(',')
            >>> pd['a2-list'] = pd['a2'].str.split(',')
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                    a1     a2  b    a1-list    a2-list
            idx
            0    1,2,3  a,b,c  1  [1, 2, 3]  [a, b, c]
            1    2,3,4  b,c,d  2  [2, 3, 4]  [b, c, d]
            2      3,4    c,d  3     [3, 4]     [c, d]
            >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
            >>> result = stack_list_columns(dd, ['a1-list', 'a2-list'], ['str', 'str'], ['b'])
            >>> print(result.compute())
            ... # doctest: +NORMALIZE_WHITESPACE
                 b a1-list a2-list
            idx
            0    1       1       a
            0    1       2       b
            0    1       3       c
            1    2       2       b
            1    2       3       c
            1    2       4       d
            2    3       3       c
            2    3       4       d
            >>> pd = pandas.DataFrame({'a1': ['1,2,3', '2,3,4', '3,4'], 'a2': ['a,b,c', 'b,c', 'c,d'],
            ...          'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
            >>> pd['a1-list'] = pd['a1'].str.split(',')
            >>> pd['a2-list'] = pd['a2'].str.split(',')
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                    a1     a2  b    a1-list    a2-list
            idx
            0    1,2,3  a,b,c  1  [1, 2, 3]  [a, b, c]
            1    2,3,4    b,c  2  [2, 3, 4]     [b, c]
            2      3,4    c,d  3     [3, 4]     [c, d]
            >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
            >>> result = stack_list_columns(dd, ['a1-list', 'a2-list'], ['str', 'str'], ['b'])
            >>> print(result.compute())
            ... # doctest: +NORMALIZE_WHITESPACE
                 b a1-list a2-list
            idx
            0    1       1       a
            0    1       2       b
            0    1       3       c
            1    2       2       b
            1    2       3       c
            1    2       4     NaN
            2    3       3       c
            2    3       4       d
        """
    meta = make_meta((dd.index.name, dd.index.dtype),
                     [(k_column, dd[k_column].dtype) for k_column in keep_columns] +
                     [(l_column, l_column_t) for l_column, l_column_t in zip(list_columns, output_dtypes)])
    return map_partitions_as_meta(dd,
                                  lambda pd: pandas_stack_list_columns(pd, list_columns, output_dtypes, keep_columns),
                                  meta)

def stack_dict_column(dd: DataFrame, dict_column: Any, label_dtype: Any, value_dtype: Any,
                      keep_columns: List[Any] = None, label_suffix: str = '_label', value_suffix: str = '_value'):
    """
    API to create stack pandas dataframe from column containing dict-like objects.

    Args:
        df (pandas.DataFrame): target dask dataframe
        dict_column (Any): name of column containing dict-like objects.
        label_dtype (Any): dtype of keys of dict column. it depends on keys of dict.
        label_dtype (Any): dtype of values of dict column. it depends on values of dict.
        keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
        label_suffix (str): label column name of stack result dataframe will be `{dict_column}{label_suffix}`
        value_suffix (str): value column name of stack result dataframe will be `{dict_column}{value_suffix}`
    Returns:
        result pandas dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [{'a': 100, 'b': 200}, {'c': 300}, {'a': 500, 'c': 1000}],
        ...                        'b': [1, 2, 3], 'idx': ['A', 'B', 'C']}).set_index('idx')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
                                 a  b
        idx
        A     {'a': 100, 'b': 200}  1
        B               {'c': 300}  2
        C    {'a': 500, 'c': 1000}  3
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> result = stack_dict_column(dd, 'a', label_dtype='str', value_dtype='int', keep_columns=['b'])
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
             b  a_value a_label
        idx
        A    1      100       a
        A    1      200       b
        B    2      300       c
        C    3      500       a
        C    3     1000       c
    """
    meta = make_meta((dd.index.name, dd.index.dtype),
                     [(k_column, dd[k_column].dtype) for k_column in keep_columns] +
                     [(f'{dict_column}{value_suffix}', value_dtype), (f'{dict_column}{label_suffix}', label_dtype)])
    return dd.map_partitions(
        lambda pd: pandas_stack_dict_column(pd, dict_column,
                                            label_dtype, value_dtype, keep_columns,
                                            label_suffix, value_suffix),
        meta=meta,
    )


def stack_columns(dd: DataFrame, target_columns: List[Any], keep_columns: List[Any],
                  label_name: Any = 'stack_label', output_name: Any = 'stacked',
                  label_dtype: Any = 'object', output_dtype: Any = 'object'):
    """
    API to create stack dask dataframe from specific columns.

    Args:
        df (pandas.DataFrame): target pandas dataframe
        target_columns (List[Any]): name of columns to stack.
        keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
        label_name (Any): name of label column in stack result
        output_name (Any): name of output column in stack result
        label_dtype (Any): dtype of label column in stack result
        output_dtype (Any): dtype of output column in stack result
    Returns:
        result dask dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> result = stack_columns(dd, ['a', 'b'], ['label'], label_dtype='object', output_dtype='int')
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
            label stack_label  stacked
        100     a           a        1
        100     a           b        2
        200     b           a        2
        200     b           b        3
        300     c           a        3
        300     c           b        4
        400     d           a        4
        400     d           b        5
        >>> print(result.dtypes)
        ... # doctest: +NORMALIZE_WHITESPACE
        label          object
        stack_label    object
        stacked         int64
        dtype: object
    """
    meta = make_meta((dd.index.name, dd.index.dtype),
                     [(k_column, dd[k_column].dtype) for k_column in keep_columns] +
                     [(label_name, label_dtype), (output_name, output_dtype)])
    return dd.map_partitions(lambda pd: pandas_stack_columns(pd, target_columns, keep_columns,
                                                             label_name, output_name, label_dtype, output_dtype),
                             meta=meta)



from typing import List, Dict, Union, Any

from dask.dataframe import DataFrame
from pandas import MultiIndex

from framania.pandasmania.aggregate import aggregate_as_series as aggregate_as_series_pandas


def flatten_aggregated_columns(dd: DataFrame):
    """
    API to make aggregated columns that are MultiIndex style flat

    Args:
        pd: target dataframe
    Returns:
        result pandas dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': range(0, 600, 100), 'key': [0, 1, 2, 0, 1, 2]})
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  key
        0    1    0    0
        1   10  100    1
        2  100  200    2
        3    1  300    0
        4    1  400    1
        5  100  500    2
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> groupby = dd.groupby(['key'])
        >>> groupby_result = groupby.agg({'a': ['sum', 'min'], 'b': ['mean', 'sum', 'max']})
        >>> print(groupby_result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
               a           b
             sum  min   mean  sum  max
        key
        0      2    1  150.0  300  300
        1     11    1  250.0  500  400
        2    200  100  350.0  700  500
        >>> flatten = flatten_aggregated_columns(groupby_result)
        >>> print(flatten.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
             a_sum  a_min  b_mean  b_sum  b_max
        key
        0        2      1   150.0    300    300
        1       11      1   250.0    500    400
        2      200    100   350.0    700    500
        >>> print(flatten_aggregated_columns(dd).compute())
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  key
        0    1    0    0
        1   10  100    1
        2  100  200    2
        3    1  300    0
        4    1  400    1
        5  100  500    2
    """
    if not isinstance(dd.columns, MultiIndex) or dd.columns.nlevels != 2:
        return dd

    result = dd.copy()

    columns = []
    for l1, l2 in zip(dd.columns.get_level_values(0), dd.columns.get_level_values(1)):
        if l2 == '':
            columns.append(l1)
        else:
            columns.append('_'.join((l1, l2)))
    result.columns = columns

    return result


def aggregate_by_named_index_and_keys(dd: DataFrame, keys: List[Any], agg: Dict[str, Union[str, List[str]]]):
    """
    API for groupby aggregation by named index and other columns.

    Args:
        df (dask.dataframe.DataFrame): target dask dataframe
        keys (List[Any]): groupby key columns that be used with named index
        agg (Dict[str, Union[str, List[str]]]): see `func` in https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.agg.html
    Returns:
        result dask dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': range(0, 600, 100), 'idx': [0, 1, 2, 0, 1, 2]}).set_index('idx')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
               a    b
        idx
        0      1    0
        1     10  100
        2    100  200
        0      1  300
        1      1  400
        2    100  500
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> result = aggregate_by_named_index_and_keys(dd, ['a'], {'b': ['mean', 'max']})
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
               a      b
                   mean  max
        idx
        0      1  150.0  300
        1      1  400.0  400
        1     10  100.0  100
        2    100  350.0  500
    """
    index_name = dd.index.name
    groupby_keys = [index_name] + keys

    lambda_agg = lambda pd: pd.groupby(by=groupby_keys).agg(agg)
    lambda_reset_index = lambda pd: pd.reset_index(keys)

    if len(keys) > 0:
        lambda_all = lambda pd: lambda_reset_index(lambda_agg(pd))
    else:
        lambda_all = lambda_agg
    return dd.map_partitions(lambda_all)


def aggregate_as_series_by_named_index_and_keys(dd: DataFrame, keys: List[Any], series: Any, agg_func: Any,
                                                output_series_name: Any):
    """
    API to output the aggregate result by named index and other columns as a column of the original data frame.

    Args:
        dd (DataFrame): target dask dataframe
        keys (List[Any): group keys other than index
        series (Any): name of series to be aggregated
        agg_func (Union[str, List[str]]): see `func` in https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.agg.html
        output_series_name (Any): name of result series
    Returns:
        result pandas series
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': range(0, 600, 100), 'idx': [0, 1, 2, 0, 1, 2]}).set_index('idx')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
               a    b
        idx
        0      1    0
        1     10  100
        2    100  200
        0      1  300
        1      1  400
        2    100  500
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> result = aggregate_as_series_by_named_index_and_keys(dd, ['a'], 'b', 'sum', 'sum_b')
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
               a    b  sum_b
        idx
        0      1    0    300
        0      1  300    300
        1     10  100    100
        1      1  400    400
        2    100  200    700
        2    100  500    700
    """
    index_name = dd.index.name
    groupby_keys = [index_name] + keys

    def assign_result_column(pd):
        pd.reset_index(drop=False, inplace=True)
        agg_result = aggregate_as_series_pandas(pd.groupby(groupby_keys), series, agg_func)
        pd[output_series_name] = agg_result
        pd.set_index(index_name, drop=True, inplace=True)
        return pd

    return dd.map_partitions(assign_result_column)


def aggregate_as_series(dd: DataFrame, keys: List[Any], series: Any, agg_func: Any,
                        output_series_name: Any):
    """
    API to output the aggregate result as a column of the original data frame.

    Args:
        dd (DataFrame): target dask dataframe
        keys (List[Any): group keys
        series (Any): name of series to be aggregated
        agg_func (Union[str, List[str]]): see `func` in https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.agg.html
        output_series_name (Any): name of result series
    Returns:
        result pandas series
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': range(0, 600, 100), 'idx': [0, 1, 2, 0, 1, 2]})
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  idx
        0    1    0    0
        1   10  100    1
        2  100  200    2
        3    1  300    0
        4    1  400    1
        5  100  500    2
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> result = aggregate_as_series(dd, ['idx', 'a'], 'b', 'sum', 'sum_b')
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  idx  sum_b
        0    1    0    0    300
        1   10  100    1    100
        2  100  200    2    700
        0    1  300    0    300
        1    1  400    1    400
        2  100  500    2    700
    """
    group_data = dd.groupby(keys)[series].agg(agg_func)
    group_data.name = output_series_name
    group_df = group_data.reset_index()
    return dd.merge(group_df, on=keys)



from pathlib import Path

from flask import Flask

from framania.hvmania.manager import HVManiaManager


class HVManiaServer(Flask):
    def __init__(self, import_name, name: str, directory: Path, renderer: str = 'bokeh', *args, **kwargs):
        super().__init__(import_name, *args, **kwargs)
        self.manager = HVManiaManager(name, directory, renderer)


def build_app(name: str, directory: Path, renderer: str = 'bokeh'):
    _app = HVManiaServer(__name__, name, directory, renderer)

    @_app.route('/hvplot/<name>')
    def chart(name):
        html_file = _app.manager[name]
        return html_file.open('rb').read()

    return _app



from collections import OrderedDict
from typing import Any, List

import numpy
import pandas
from pandas import DataFrame
from pandas.core.indexes.frozen import FrozenList


def stack_list_column(pd: DataFrame, list_column: Any, output_dtype: Any, keep_columns: List[Any] = None):
    """
    API to create stack pandas dataframe from column containing list-like objects.

    Args:
        df (pandas.DataFrame): target pandas dataframe
        list_column (Any): name of column containing list-like objects.
        output_dtype (Any): dtype of output column. may depends on contents of list.
        keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
    Returns:
        result pandas dataframe
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': ['1,2,3', '2,3,4', '3,4'], 'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
        >>> pd['a-list'] = pd['a'].str.split(',')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
                 a  b     a-list
        idx
        0    1,2,3  1  [1, 2, 3]
        1    2,3,4  2  [2, 3, 4]
        2      3,4  3     [3, 4]
        >>> result = stack_list_column(pd, 'a-list', 'str', ['b'])
        >>> print(result)
        ... # doctest: +NORMALIZE_WHITESPACE
             b a-list
        idx
        0    1      1
        0    1      2
        0    1      3
        1    2      2
        1    2      3
        1    2      4
        2    3      3
        2    3      4
    """

    ipd = pd.set_index(keep_columns, append=True)

    index = []
    values = []

    # TODO: more efficient stack implementation if exists
    for idx, row in ipd[list_column].iteritems():
        index.extend([idx] * len(row))
        values.extend(row)
    ix = pandas.Index(index, names=ipd.index.names)
    s = pandas.Series(values, dtype=output_dtype, index=ix)
    result = pandas.DataFrame({list_column: s})
    result.reset_index(keep_columns, drop=False, inplace=True)
    return result


def stack_list_columns(pd: DataFrame, list_columns: List[Any], output_dtypes: List[Any],
                       keep_columns: List[Any] = None):
    """
        API to create stack pandas dataframe from columns containing list-like objects.
        List elements which has same list index will be assigned to same row.
        If length of lists are not same, NaN will be filled automatically.

        Args:
            df (pandas.DataFrame): target pandas dataframe
            list_columns (List[Any]): name of columns containing list-like objects.
            output_dtypes (List[Any]): dtypes of output columns. may depends on contents of list.
            keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
        Returns:
            result pandas dataframe
        Examples:
            >>> import pandas
            >>> pd = pandas.DataFrame({'a1': ['1,2,3', '2,3,4', '3,4'], 'a2': ['a,b,c', 'b,c,d', 'c,d'],
            ...          'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
            >>> pd['a1-list'] = pd['a1'].str.split(',')
            >>> pd['a2-list'] = pd['a2'].str.split(',')
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                    a1     a2  b    a1-list    a2-list
            idx
            0    1,2,3  a,b,c  1  [1, 2, 3]  [a, b, c]
            1    2,3,4  b,c,d  2  [2, 3, 4]  [b, c, d]
            2      3,4    c,d  3     [3, 4]     [c, d]
            >>> result = stack_list_columns(pd, ['a1-list', 'a2-list'], ['str', 'str'], ['b'])
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
                 b a1-list a2-list
            idx
            0    1       1       a
            0    1       2       b
            0    1       3       c
            1    2       2       b
            1    2       3       c
            1    2       4       d
            2    3       3       c
            2    3       4       d
            >>> pd = pandas.DataFrame({'a1': ['1,2,3', '2,3,4', '3,4'], 'a2': ['a,b,c', 'b,c', 'c,d'],
            ...          'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
            >>> pd['a1-list'] = pd['a1'].str.split(',')
            >>> pd['a2-list'] = pd['a2'].str.split(',')
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                    a1     a2  b    a1-list    a2-list
            idx
            0    1,2,3  a,b,c  1  [1, 2, 3]  [a, b, c]
            1    2,3,4    b,c  2  [2, 3, 4]     [b, c]
            2      3,4    c,d  3     [3, 4]     [c, d]
            >>> result = stack_list_columns(pd, ['a1-list', 'a2-list'], ['str', 'str'], ['b'])
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
                 b a1-list a2-list
            idx
            0    1       1       a
            0    1       2       b
            0    1       3       c
            1    2       2       b
            1    2       3       c
            1    2       4     NaN
            2    3       3       c
            2    3       4       d
        """
    ipd = pd.set_index(keep_columns, append=True)

    index = []
    values = [[] for _ in list_columns]

    # TODO: more efficient stack implementation if exists
    for idx, row in ipd.iterrows():
        list_column_values = [row[list_column] for list_column in list_columns]
        max_len = max(len(lcv) for lcv in list_column_values)
        index.extend([idx] * max_len)
        for vs, _list_column_values in zip(values, list_column_values):
            vlen = len(_list_column_values)
            none_values = [numpy.nan] * (max_len - vlen)
            vs.extend(_list_column_values)
            vs.extend(none_values)
        values.extend(row)
    ix = pandas.Index(index, names=ipd.index.names)
    series_s = [pandas.Series(vs, dtype=output_dtype, index=ix) for vs, output_dtype in zip(values, output_dtypes)]
    result = pandas.DataFrame(OrderedDict([(list_column, s) for list_column, s in zip(list_columns, series_s)]))
    result.reset_index(keep_columns, drop=False, inplace=True)
    return result


def stack_dict_column(pd: DataFrame, dict_column: Any, label_dtype: Any, value_dtype: Any,
                      keep_columns: List[Any] = None, label_suffix: str = '_label', value_suffix: str = '_value'):
    """
    API to create stack pandas dataframe from column containing dict-like objects.

    Args:
        df (pandas.DataFrame): target pandas dataframe
        dict_column (Any): name of column containing dict-like objects.
        label_dtype (Any): dtype of keys of dict column. it depends on keys of dict.
        label_dtype (Any): dtype of values of dict column. it depends on values of dict.
        keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
        label_suffix (str): label column name of stack result dataframe will be `{dict_column}{label_suffix}`
        value_suffix (str): value column name of stack result dataframe will be `{dict_column}{value_suffix}`
    Returns:
        result pandas dataframe
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [{'a': 100, 'b': 200}, {'c': 300}, {'a': 500, 'c': 1000}],
        ...                        'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
                                 a  b
        idx
        0     {'a': 100, 'b': 200}  1
        1               {'c': 300}  2
        2    {'a': 500, 'c': 1000}  3
        >>> result = stack_dict_column(pd, 'a', label_dtype='str', value_dtype='int', keep_columns=['b'])
        >>> print(result)
        ... # doctest: +NORMALIZE_WHITESPACE
             b  a_value a_label
        idx
        0    1      100       a
        0    1      200       b
        1    2      300       c
        2    3      500       a
        2    3     1000       c
        >>> print(result.dtypes)
        ... # doctest: +NORMALIZE_WHITESPACE
        b           int64
        a_value     int64
        a_label    object
        dtype: object
    """

    ipd = pd.set_index(keep_columns, append=True)

    index = []
    labels = []
    values = []

    # TODO: more efficient stack implementation if exists
    for idx, row in ipd[dict_column].iteritems():
        index.extend([idx] * len(row))
        for l, v in row.items():
            values.append(v)
            labels.append(l)
    ix = pandas.Index(index, names=ipd.index.names)
    ls = pandas.Series(labels, dtype=label_dtype, index=ix)
    vs = pandas.Series(values, dtype=value_dtype, index=ix)
    result = pandas.DataFrame(OrderedDict([(f'{dict_column}{value_suffix}', vs),
                                           (f'{dict_column}{label_suffix}', ls)]))
    result.reset_index(keep_columns, drop=False, inplace=True)
    return result


def stack_columns(pd: DataFrame, target_columns: List[Any], keep_columns: List[Any],
                  label_name: Any = 'stack_label', output_name: Any = 'stacked',
                  label_dtype: Any = 'object', output_dtype: Any = 'object'):
    """
    API to create stack pandas dataframe from specific columns.

    Args:
        df (pandas.DataFrame): target pandas dataframe
        target_columns (List[Any]): name of columns to stack.
        keep_columns (List[Any]): result dataframe will contains original index, stacked column, and keep_columns
        label_name (Any): name of label column in stack result
        output_name (Any): name of output column in stack result
        label_dtype (Any): dtype of label column in stack result
        output_dtype (Any): dtype of output column in stack result
    Returns:
        result pandas dataframe
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> result = stack_columns(pd, ['a', 'b'], ['label'])
        >>> print(result)
        ... # doctest: +NORMALIZE_WHITESPACE
            label stack_label  stacked
        100     a           a        1
        100     a           b        2
        200     b           a        2
        200     b           b        3
        300     c           a        3
        300     c           b        4
        400     d           a        4
        400     d           b        5
        >>> print(result.dtypes)
        ... # doctest: +NORMALIZE_WHITESPACE
        label          object
        stack_label    object
        stacked        object
        dtype: object
    """
    indexed = pd.set_index(keep_columns, append=True)
    stacked = indexed[target_columns].stack()
    new_index_names = list(stacked.index.names)
    new_index_names[-1] = label_name
    renamed = stacked.rename_axis(index=FrozenList(new_index_names))
    renamed.name = output_name

    result_df = DataFrame(renamed).reset_index(level=keep_columns + [label_name])

    result_df[label_name] = result_df[label_name].astype(label_dtype)
    result_df[output_name] = result_df[output_name].astype(output_dtype)

    return result_df


def unstack_column(pd: DataFrame, unstack_columns: List[Any], on: List[Any], value_column: Any, splitter: str = '_',
                   suffix: str = None):
    """
    API to create unstack pandas dataframe from specific columns.

    Args:
        df (pandas.DataFrame): target pandas dataframe
        unstack_columns (List[Any]): name of columns to be used as transposing dimension.
        on (List[Any]): name of columns to be kept as row dimensions.
        label_name (Any): name of label column in stack result
        value_column (Any): name of column which contains values to be transpose.
        splitter (str): spacer to merge multiple transposing dimension values.
        suffix (str): suffix string for transposing dimension columns
    Returns:
        result pandas dataframe
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> result = unstack_column(pd, ['a', 'b'], ['label'], 'c')
        >>> print(result)
        ... # doctest: +NORMALIZE_WHITESPACE
          label  1_2c  2_3c  3_4c  4_5c
        0     a   5.0   NaN   NaN   NaN
        1     b   NaN   6.0   NaN   NaN
        2     c   NaN   NaN   7.0   NaN
        3     d   NaN   NaN   NaN   8.0
        >>> print(result.dtypes)
        ... # doctest: +NORMALIZE_WHITESPACE
        label     object
        1_2c     float64
        2_3c     float64
        3_4c     float64
        4_5c     float64
        dtype: object
    """
    index_df = pd.set_index(on + unstack_columns)
    result = index_df[[value_column]].unstack(level=unstack_columns)
    column_len = len(result.columns)
    column_names = []
    if suffix is None:
        suffix = value_column
    for values_i in range(column_len):
        unstack_column_values = \
            splitter.join([str(result.columns.get_level_values(c)[values_i]) for c in unstack_columns]) + suffix
        column_names.append(unstack_column_values)
    result.columns = column_names
    return result.reset_index(level=on)
